@@ -33,6 +33,11 @@
 #error Must have C++11 or newer.
 #endif
 
+// windows.h can conflict with Criterion::STRICT
+#ifdef STRICT
+#undef STRICT
+#endif
+
 #include <exception>
 #include <map>
 #include <memory>
@@ -50,10 +55,12 @@ namespace proj {}
 //! @cond Doxygen_Suppress
 
 #ifndef PROJ_DLL
+#if defined(_MSC_VER)
 #ifdef PROJ_MSVC_DLL_EXPORT
 #define PROJ_DLL __declspec(dllexport)
-#elif defined(PROJ_MSVC_DLL_IMPORT)
+#else
 #define PROJ_DLL __declspec(dllimport)
+#endif
 #elif defined(__GNUC__)
 #define PROJ_DLL __attribute__((visibility("default")))
 #else
@@ -62,12 +69,7 @@ namespace proj {}
 #endif
 
 #ifndef PROJ_MSVC_DLL
-
-#ifdef PROJ_MSVC_DLL_EXPORT
-#define PROJ_MSVC_DLL PROJ_DLL
-#define PROJ_GCC_DLL
-#define PROJ_INTERNAL
-#elif defined(PROJ_MSVC_DLL_IMPORT)
+#if defined(_MSC_VER)
 #define PROJ_MSVC_DLL PROJ_DLL
 #define PROJ_GCC_DLL
 #define PROJ_INTERNAL
@@ -84,14 +86,21 @@ namespace proj {}
 #define PROJ_GCC_DLL
 #define PROJ_INTERNAL
 #endif
-
 #define PROJ_FOR_TEST PROJ_DLL
-
 #endif
 
 #include "nn.hpp"
 
 /* To allow customizing the base namespace of PROJ */
+#ifdef PROJ_INTERNAL_CPP_NAMESPACE
+#define NS_PROJ osgeo::internalproj
+#define NS_PROJ_START                                                          \
+    namespace osgeo {                                                          \
+    namespace internalproj {
+#define NS_PROJ_END                                                            \
+    }                                                                          \
+    }
+#else
 #ifndef NS_PROJ
 #define NS_PROJ osgeo::proj
 #define NS_PROJ_START                                                          \
@@ -101,6 +110,7 @@ namespace proj {}
     }                                                                          \
     }
 #endif
+#endif
 
 // Private-implementation (Pimpl) pattern
 #define PROJ_OPAQUE_PRIVATE_DATA                                               \
@@ -120,11 +130,11 @@ namespace proj {}
 // to be able to call make_shared on a protected/private constructor
 #define INLINED_MAKE_SHARED                                                    \
     template <typename T, typename... Args>                                    \
-    static std::shared_ptr<T> make_shared(Args &&... args) {                   \
+    static std::shared_ptr<T> make_shared(Args &&...args) {                    \
         return std::shared_ptr<T>(new T(std::forward<Args>(args)...));         \
     }                                                                          \
     template <typename T, typename... Args>                                    \
-    static util::nn_shared_ptr<T> nn_make_shared(Args &&... args) {            \
+    static util::nn_shared_ptr<T> nn_make_shared(Args &&...args) {             \
         return util::nn_shared_ptr<T>(                                         \
             util::i_promise_i_checked_for_null,                                \
             std::shared_ptr<T>(new T(std::forward<Args>(args)...)));           \
@@ -134,7 +144,7 @@ namespace proj {}
 // to be able to call make_unique on a protected/private constructor
 #define INLINED_MAKE_UNIQUE                                                    \
     template <typename T, typename... Args>                                    \
-    static std::unique_ptr<T> make_unique(Args &&... args) {                   \
+    static std::unique_ptr<T> make_unique(Args &&...args) {                    \
         return std::unique_ptr<T>(new T(std::forward<Args>(args)...));         \
     }
 
@@ -166,6 +176,13 @@ namespace proj {}
 
 NS_PROJ_START
 
+//! @cond Doxygen_Suppress
+namespace io {
+class DatabaseContext;
+using DatabaseContextPtr = std::shared_ptr<DatabaseContext>;
+} // namespace io
+//! @endcond
+
 /** osgeo.proj.util namespace.
  *
  * \brief A set of base types from ISO 19103, \ref GeoAPI and other PROJ
@@ -196,6 +213,14 @@ template <typename T> using nn_shared_ptr = nn<std::shared_ptr<T>>;
 // To avoid formatting differences between clang-format 3.8 and 7
 #define PROJ_NOEXCEPT noexcept
 
+//! @cond Doxygen_Suppress
+// isOfExactType<MyType>(*p) checks that the type of *p is exactly MyType
+template <typename TemplateT, typename ObjectT>
+inline bool isOfExactType(const ObjectT &o) {
+    return typeid(TemplateT).hash_code() == typeid(o).hash_code();
+}
+//! @endcond
+
 /** \brief Loose transposition of [std::optional]
  * (https://en.cppreference.com/w/cpp/utility/optional) available from C++17. */
 template <class T> class optional {
@@ -206,7 +231,7 @@ template <class T> class optional {
     inline explicit optional(T &&val)
         : hasVal_(true), val_(std::forward<T>(val)) {}
 
-    inline optional(const optional &other) = default;
+    inline optional(const optional &) = default;
     inline optional(optional &&other) PROJ_NOEXCEPT
         : hasVal_(other.hasVal_),
           // cppcheck-suppress functionStatic
@@ -224,7 +249,7 @@ template <class T> class optional {
         val_ = std::forward<T>(val);
         return *this;
     }
-    inline optional &operator=(const optional &other) = default;
+    inline optional &operator=(const optional &) = default;
     inline optional &operator=(optional &&other) noexcept {
         hasVal_ = other.hasVal_;
         val_ = std::forward<T>(other.val_);
@@ -289,7 +314,7 @@ struct BaseObjectNNPtr : public util::nn<BaseObjectPtr> {
 using BaseObjectNNPtr = util::nn<BaseObjectPtr>;
 #endif
 
-/** \brief Class that can be derived from, to emulate Java's Object behaviour.
+/** \brief Class that can be derived from, to emulate Java's Object behavior.
  */
 class PROJ_GCC_DLL BaseObject {
   public:
@@ -306,6 +331,7 @@ class PROJ_GCC_DLL BaseObject {
   protected:
     PROJ_INTERNAL BaseObject();
     PROJ_INTERNAL void assignSelf(const BaseObjectNNPtr &self);
+    PROJ_INTERNAL BaseObject &operator=(BaseObject &&other);
 
   private:
     PROJ_OPAQUE_PRIVATE_DATA
@@ -342,15 +368,18 @@ class PROJ_GCC_DLL IComparable {
         EQUIVALENT_EXCEPT_AXIS_ORDER_GEOGCRS,
     };
 
-    PROJ_DLL bool isEquivalentTo(const IComparable *other,
-                                 Criterion criterion = Criterion::STRICT) const;
+    PROJ_DLL bool
+    isEquivalentTo(const IComparable *other,
+                   Criterion criterion = Criterion::STRICT,
+                   const io::DatabaseContextPtr &dbContext = nullptr) const;
 
     PROJ_PRIVATE :
 
         //! @cond Doxygen_Suppress
         PROJ_INTERNAL virtual bool
-        _isEquivalentTo(const IComparable *other,
-                        Criterion criterion = Criterion::STRICT) const = 0;
+        _isEquivalentTo(
+            const IComparable *other, Criterion criterion = Criterion::STRICT,
+            const io::DatabaseContextPtr &dbContext = nullptr) const = 0;
     //! @endcond
 };
 
@@ -672,7 +701,7 @@ class CodeList {
     //! @endcond
   protected:
     explicit CodeList(const std::string &nameIn) : name_(nameIn) {}
-    CodeList(const CodeList &other) = default;
+    CodeList(const CodeList &) = default;
     CodeList &operator=(const CodeList &other);
 
   private:
